@@ -21,13 +21,12 @@ import array
 import contextlib
 import inspect
 import traceback
-import warnings
 
 import xcffib.xproto
 from xcffib.xproto import EventMask, SetMode, StackMode
 
 from libqtile import hook, utils
-from libqtile.command_object import CommandError, CommandObject
+from libqtile.command.base import CommandError, CommandObject
 from libqtile.log_utils import logger
 
 # ICCM Constants
@@ -176,7 +175,6 @@ class _Window(CommandObject):
         self.name = "<no name>"
         self.strut = None
         self.state = NormalState
-        self.window_type = "normal"
         self._float_state = NOT_FLOATING
         self._demands_attention = False
 
@@ -199,22 +197,14 @@ class _Window(CommandObject):
 
     x = property(fset=_geometry_setter("x"), fget=_geometry_getter("x"))
     y = property(fset=_geometry_setter("y"), fget=_geometry_getter("y"))
-
-    @property
-    def width(self):
-        return _geometry_getter("width")(self)
-
-    @width.setter
-    def width(self, value):
-        _geometry_setter("width")(self, value)
-
-    @property
-    def height(self):
-        return _geometry_getter("height")(self)
-
-    @height.setter
-    def height(self, value):
-        _geometry_setter("height")(self, value)
+    width = property(
+        fset=_geometry_setter("width"),
+        fget=_geometry_getter("width"),
+    )
+    height = property(
+        fset=_geometry_setter("height"),
+        fget=_geometry_getter("height"),
+    )
 
     float_x = property(
         fset=_float_setter("x"),
@@ -237,6 +227,17 @@ class _Window(CommandObject):
     def has_focus(self):
         return self == self.qtile.current_window
 
+    def has_fixed_size(self):
+        try:
+            if ('PMinSize' in self.hints['flags'] and
+                    'PMaxSize' in self.hints['flags'] and
+                    0 < self.hints["min_width"] == self.hints["max_width"] and
+                    0 < self.hints["min_height"] == self.hints["max_height"]):
+                return True
+        except KeyError:
+            pass
+        return False
+
     def has_user_set_position(self):
         try:
             if 'USPosition' in self.hints['flags'] or 'PPosition' in self.hints['flags']:
@@ -263,42 +264,7 @@ class _Window(CommandObject):
         except (xcffib.xproto.WindowError, xcffib.xproto.AccessError):
             return
 
-        # FIXME
-        # h values
-        # {
-        #    'icon_pixmap': 4194337,
-        #    'icon_window': 0,
-        #    'icon_mask': 4194340,
-        #    'icon_y': 0,
-        #    'input': 1,
-        #    'icon_x': 0,
-        #    'window_group': 4194305
-        #    'initial_state': 1,
-        #    'flags': set(['StateHint',
-        #                  'IconMaskHint',
-        #                  'WindowGroupHint',
-        #                  'InputHint',
-        #                  'UrgencyHint',
-        #                  'IconPixmapHint']),
-        # }
-
         if normh:
-            normh['min_width'] = max(0, normh.get('min_width', 0))
-            normh['min_height'] = max(0, normh.get('min_height', 0))
-            if not normh['base_width'] and \
-                    normh['min_width'] and \
-                    normh['width_inc']:
-                # seems xcffib does ignore base width :(
-                normh['base_width'] = (
-                    normh['min_width'] % normh['width_inc']
-                )
-            if not normh['base_height'] and \
-                    normh['min_height'] and \
-                    normh['height_inc']:
-                # seems xcffib does ignore base height :(
-                normh['base_height'] = (
-                    normh['min_height'] % normh['height_inc']
-                )
             self.hints.update(normh)
 
         if h and 'UrgencyHint' in h['flags']:
@@ -418,6 +384,7 @@ class _Window(CommandObject):
         # We don't want to get the UnmapNotify for this unmap
         with self.disable_mask(xcffib.xproto.EventMask.StructureNotify):
             self.window.unmap()
+        self.state = IconicState
         self.hidden = True
 
     def unhide(self):
@@ -443,7 +410,20 @@ class _Window(CommandObject):
 
     def place(self, x, y, width, height, borderwidth, bordercolor,
               above=False, margin=None):
-        """Places the window at the specified location with the given size.
+        """
+        Places the window at the specified location with the given size.
+
+        Parameters
+        ==========
+        x : int
+        y : int
+        width : int
+        height : int
+        borderwidth : int
+        bordercolor : string
+        above : bool, optional
+        margin : int or list, optional
+            space around window as int or list of ints [N E S W]
         """
 
         # TODO: self.x/y/height/width are updated BEFORE
@@ -462,10 +442,12 @@ class _Window(CommandObject):
 
         # Adjust the placement to account for layout margins, if there are any.
         if margin is not None:
-            x += margin
-            y += margin
-            width -= margin * 2
-            height -= margin * 2
+            if isinstance(margin, int):
+                margin = [margin] * 4
+            x += margin[3]
+            y += margin[0]
+            width -= margin[1] + margin[3]
+            height -= margin[0] + margin[2]
 
         # save x and y float offset
         if self.group is not None and self.group.screen is not None:
@@ -476,33 +458,27 @@ class _Window(CommandObject):
         self.y = y
         self.width = width
         self.height = height
-        self.borderwidth = borderwidth
-        self.bordercolor = bordercolor
 
         kwarg = dict(
             x=x,
             y=y,
             width=width,
             height=height,
-            borderwidth=borderwidth,
         )
         if above:
             kwarg['stackmode'] = StackMode.Above
 
         self.window.configure(**kwarg)
+        self.paint_borders(bordercolor, borderwidth)
 
         if send_notify:
             self.send_configure_notify(x, y, width, height)
 
-        if bordercolor is not None:
-            self.window.set_attribute(borderpixel=bordercolor)
-
-    def user_placed_window_setup(self, borderpixel, borderwidth):
+    def paint_borders(self, borderpixel, borderwidth):
         self.borderwidth = borderwidth
         self.bordercolor = borderpixel
-        self.window.set_attribute(borderpixel=borderpixel)
-        self.window.configure(borderwidth=borderwidth, stackmode=StackMode.Above)
-        self.unhide()
+        self.window.configure(borderwidth=borderwidth)
+        self.window.paint_borders(borderpixel)
 
     def send_configure_notify(self, x, y, width, height):
         """Send a synthetic ConfigureNotify"""
@@ -715,13 +691,14 @@ class Static(_Window):
         self.conf_y = y
         self.conf_width = width
         self.conf_height = height
-        self.x = x or 0
-        self.y = y or 0
+        x = x or 0
+        y = y or 0
+        self.x = x + screen.x
+        self.y = y + screen.y
         self.width = width or 0
         self.height = height or 0
         self.screen = screen
-        if None not in (x, y, width, height):
-            self.place(x, y, width, height, 0, 0)
+        self.place(self.x, self.y, self.width, self.height, 0, 0)
         self.update_strut()
 
     def handle_ConfigureRequest(self, e):  # noqa: N802
@@ -846,18 +823,6 @@ class Window(_Window):
     def toggle_floating(self):
         self.floating = not self.floating
 
-    def togglefloating(self):
-        warnings.warn("togglefloating is deprecated, use toggle_floating", DeprecationWarning)
-        self.toggle_floating()
-
-    def enablefloating(self):
-        warnings.warn("enablefloating is deprecated, use floating=True", DeprecationWarning)
-        self.floating = True
-
-    def disablefloating(self):
-        warnings.warn("disablefloating is deprecated, use floating=False", DeprecationWarning)
-        self.floating = False
-
     @property
     def fullscreen(self):
         return self._float_state == FULLSCREEN
@@ -895,10 +860,6 @@ class Window(_Window):
     def toggle_fullscreen(self):
         self.fullscreen = not self.fullscreen
 
-    def togglefullscreen(self):
-        warnings.warn("togglefullscreen is deprecated, use toggle_fullscreen", DeprecationWarning)
-        self.toggle_fullscreen()
-
     @property
     def maximized(self):
         return self._float_state == MAXIMIZED
@@ -920,17 +881,9 @@ class Window(_Window):
             if self._float_state == MAXIMIZED:
                 self.floating = False
 
-    def enablemaximize(self, state=MAXIMIZED):
-        warnings.warn("enablemaximize is deprecated, use maximized=True", DeprecationWarning)
-        self.maximized = True
-
     def toggle_maximize(self, state=MAXIMIZED):
         self.maximized = not self.maximized
 
-    def togglemaximize(self):
-        warnings.warn("togglemaximize is deprecated, use toggle_maximize", DeprecationWarning)
-        self.toggle_maximize()
-
     @property
     def minimized(self):
         return self._float_state == MINIMIZED
@@ -944,18 +897,10 @@ class Window(_Window):
             if self._float_state == MINIMIZED:
                 self.floating = False
 
-    def enableminimize(self):
-        warnings.warn("enableminimized is deprecated, use minimized=True", DeprecationWarning)
-        self.minimized = True
-
     def toggle_minimize(self):
         self.minimized = not self.minimized
 
-    def toggleminimize(self):
-        warnings.warn("toggleminimize is deprecated, use toggle_minimize", DeprecationWarning)
-        self.toggle_minimize()
-
-    def cmd_static(self, screen, x=None, y=None, width=None, height=None):
+    def cmd_static(self, screen=None, x=None, y=None, width=None, height=None):
         """Makes this window a static window, attached to a Screen
 
         If any of the arguments are left unspecified, the values given by the
@@ -964,7 +909,10 @@ class Window(_Window):
         anything.
         """
         self.defunct = True
-        screen = self.qtile.screens[screen]
+        if screen is None:
+            screen = self.qtile.current_screen
+        else:
+            screen = self.qtile.screens[screen]
         if self.group:
             self.group.remove(self)
         s = Static(self.window, self.qtile, screen, x, y, width, height)
@@ -994,7 +942,9 @@ class Window(_Window):
         if self.width < 0:
             self.width = 0
 
-        screen = self.qtile.find_closest_screen(self.x, self.y)
+        screen = self.qtile.find_closest_screen(
+            self.x + self.width // 2, self.y + self.height // 2
+        )
         if self.group and screen is not None and screen != self.group.screen:
             self.group.remove(self, force=True)
             screen.group.add(self, force=True)
@@ -1010,11 +960,18 @@ class Window(_Window):
 
     def _reconfigure_floating(self, new_float_state=FLOATING):
         if new_float_state == MINIMIZED:
-            self.state = IconicState
             self.hide()
         else:
-            width = max(self.width, self.hints.get('min_width', 0))
-            height = max(self.height, self.hints.get('min_height', 0))
+            width = self.width
+            height = self.height
+
+            flags = self.hints.get("flags", {})
+            if "PMinSize" in flags:
+                width = max(self.width, self.hints.get('min_width', 0))
+                height = max(self.height, self.hints.get('min_height', 0))
+            if "PMaxSize" in flags:
+                width = min(width, self.hints.get('max_width', 0)) or width
+                height = min(height, self.hints.get('max_height', 0)) or height
 
             if self.hints['base_width'] and self.hints['width_inc']:
                 width_adjustment = (width - self.hints['base_width']) % self.hints['width_inc']
@@ -1087,48 +1044,26 @@ class Window(_Window):
                 raise CommandError('No such screen: %d' % index)
         self.togroup(screen.group.name)
 
-    def match(self, wname=None, wmclass=None, role=None):
+    def match(self, match):
         """Match window against given attributes.
 
         Parameters
         ==========
-        wname :
-            matches against the window name or title, that is, either
-            ``_NET_WM_VISIBLE_NAME``, ``_NET_WM_NAME``, ``WM_NAME``.
-        wmclass :
-            matches against any of the two values in the ``WM_CLASS`` property
-        role :
-            matches against the ``WM_WINDOW_ROLE`` property
+        match:
+            a config.Match object
         """
-        if not (wname or wmclass or role):
-            raise TypeError(
-                "Either a name, a wmclass or a role must be specified"
-            )
-        if wname and wname == self.name:
-            return True
-
         try:
-            cliclass = self.window.get_wm_class()
-            if wmclass and cliclass and wmclass in cliclass:
-                return True
-
-            clirole = self.window.get_wm_window_role()
-            if role and clirole and role == clirole:
-                return True
+            return match.compare(self)
         except (xcffib.xproto.WindowError, xcffib.xproto.AccessError):
             return False
 
-        return False
-
     def handle_EnterNotify(self, e):  # noqa: N802
         hook.fire("client_mouse_enter", self)
-        if self.qtile.config.follow_mouse_focus and \
-                self.group.current_window != self:
-            self.group.focus(self, False)
-        if self.group.screen and \
-                self.qtile.current_screen != self.group.screen and \
-                self.qtile.config.follow_mouse_focus:
-            self.qtile.focus_screen(self.group.screen.index, False)
+        if self.qtile.config.follow_mouse_focus:
+            if self.group.current_window != self:
+                self.group.focus(self, False)
+            if self.group.screen and self.qtile.current_screen != self.group.screen:
+                self.qtile.focus_screen(self.group.screen.index, False)
         return True
 
     def handle_ConfigureRequest(self, e):  # noqa: N802
@@ -1240,6 +1175,16 @@ class Window(_Window):
                     logger.info("Ignoring focus request")
                 else:
                     logger.warning("Invalid value for focus_on_window_activation: {}".format(focus_behavior))
+        elif atoms["_NET_CLOSE_WINDOW"] == opcode:
+            self.kill()
+        elif atoms["WM_CHANGE_STATE"] == opcode:
+            state = data.data32[0]
+            if state == NormalState:
+                self.minimized = False
+            elif state == IconicState:
+                self.minimized = True
+        else:
+            logger.info("Unhandled client message: %s", atoms.get_name(opcode))
 
     def handle_PropertyNotify(self, e):  # noqa: N802
         name = self.qtile.conn.atoms.get_name(e.atom)
@@ -1373,6 +1318,11 @@ class Window(_Window):
         """Set window dimensions to w and h"""
         self.tweak_float(w=w, h=h)
 
+    def cmd_place(self, x, y, width, height, borderwidth, bordercolor,
+                  above=False, margin=None):
+        self.place(x, y, width, height, borderwidth, bordercolor, above,
+                   margin)
+
     def cmd_get_position(self):
         return self.getposition()
 
@@ -1391,12 +1341,6 @@ class Window(_Window):
     def cmd_toggle_maximize(self):
         self.toggle_maximize()
 
-    def cmd_enable_maximize(self):
-        self.maximize = True
-
-    def cmd_disable_maximize(self):
-        self.maximize = False
-
     def cmd_toggle_fullscreen(self):
         self.toggle_fullscreen()
 
@@ -1409,12 +1353,6 @@ class Window(_Window):
     def cmd_toggle_minimize(self):
         self.toggle_minimize()
 
-    def cmd_enable_minimize(self):
-        self.minimize = True
-
-    def cmd_disable_minimize(self):
-        self.minimize = False
-
     def cmd_bring_to_front(self):
         if self.floating:
             self.window.configure(stackmode=StackMode.Above)
@@ -1449,9 +1387,9 @@ class Window(_Window):
         return (window.edges[0] <= x <= window.edges[2] and
                 window.edges[1] <= y <= window.edges[3])
 
-    def cmd_set_position(self, dx, dy):
+    def cmd_set_position(self, x, y):
         if self.floating:
-            self.tweak_float(dx, dy)
+            self.tweak_float(x, y)
             return
         for window in self.group.windows:
             if window == self or window.floating:
